'''Distribution evaluators that are usually called proportional.

This contains the most common distribution evaluators used in party-list
elections that at least aim to be proportional (although some parameter setups
might make the result very nonproportional, such as the Imperiali divisor)
and also some similar, simpler distribution evaluators that aim for strict
proportionality while relaxing some of the constraints of a distribution
evaluator.
'''

import bisect
from fractions import Fraction
from typing import Dict, Union, Callable
from numbers import Number

from .. import util
from ..candidate import Candidate
from ..component import quota, divisor
from . import core


INF = float('inf')


class PureProportionality:
    '''Distribute seats among candidates strictly proportionally (no rounding).

    This evaluator is mostly auxiliary since it gives fractional seat counts.
    '''
    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, Fraction]:
        '''Distribute seats exactly proportionally, giving fractional seats.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the proportional result
            awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        fixed = []
        result = {}
        prev_len_fixed = -1
        while len(fixed) != prev_len_fixed:
            prev_len_fixed = len(fixed)
            result = {
                cand: cseats for cand, cseats in result.items()
                if cand in fixed
            }
            budget = n_seats - sum(result.values())
            current_votes = {
                cand: n_votes for cand, n_votes in votes.items()
                if cand not in fixed
            }
            seats_per_vote = Fraction(budget, sum(current_votes.values()))
            for cand, n_votes in current_votes.items():
                cand_give_seats = n_votes * seats_per_vote
                if int(cand_give_seats) == cand_give_seats:
                    cand_give_seats = int(cand_give_seats)
                cand_has_seats = prev_gains.get(cand, 0)
                cand_max_seats = max_seats.get(cand, INF)
                if cand_give_seats > cand_has_seats:
                    if cand_give_seats > cand_max_seats:
                        # proportional result over maximum, fix maximum
                        fixed.append(cand)
                        cand_give_seats = cand_max_seats
                    result[cand] = cand_give_seats
                else:
                    # proportional result below minimum, fix minimum
                    fixed.append(cand)
                    result[cand] = cand_has_seats
        return {
            cand: cseats - prev_gains.get(cand, 0)
            for cand, cseats in result.items()
        }


class QuotaDistributor:
    '''Distribute seats proportionally, according to multiples of quota filled.

    Each contestant is awarded the number of seats according to the number
    of times their votes fill the provided quota. (This essentially means the
    numbers of votes are divided by the quota and rounded down to get the
    numbers of seats.)

    This is usually not a self-standing evaluator because it (except for very
    rare cases) does not award the full number of seats.

    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the :mod:`quota` module.
    '''
    def __init__(self,
                 quota_function: Union[
                     str, Callable[[int, int], Number]
                 ] = 'droop',
                 ):
        self.quota_function = quota.construct(quota_function)

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute seats proportionally by multiples of quota filled.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the proportional result
            awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        quota_number = self.quota_function(
            sum(votes.values()), n_seats
        )
        selected = {}
        n_overshot = 0
        overshot_candidates = []
        for candidate, n_votes in votes.items():
            n_prev = prev_gains.get(candidate, 0)
            if n_votes >= quota_number:
                n_add_seats = int(Fraction(n_votes, quota_number)) - n_prev
                if n_add_seats > 0:
                    if n_add_seats + n_prev > max_seats.get(candidate, INF):
                        overshoot = n_add_seats + n_prev
                        n_add_seats -= overshoot
                        n_overshot += overshoot
                        overshot_candidates.append(candidate)
                    selected[candidate] = n_add_seats
        if n_overshot:
            remaining_votes = {
                cand: n_votes for cand in votes.keys()
                if cand not in overshot_candidates
            }
            total_gained = {
                cand: selected.get(cand, 0) + prev_gains.get(cand, 0)
                for cand in votes
            }
            util.add_dict_to_dict(selected, self.evaluate(
                remaining_votes,
                n_overshot,
                prev_gains=total_gained,
                max_seats=max_seats
            ))
        return selected


class LargestRemainder:
    '''Distribute seats proportionally, rounding by largest remainder.

    Each contestant is awarded the number of seats according to the number
    of times their votes fill the provided quota, just like
    :class:`QuotaDistributor`. In addition to that, the
    parties that have the largest remainder of votes after the quotas are
    subtracted get an extra seat until the number of seats is exhausted.

    This includes some popular proportional party-list systems like Hare,
    Droop or Hagenbach-Bischoff. The result is usually very close to
    proportionality (closer than in highest averages systems) but might suffer
    from an Alabama paradox where adding proportionally distributed votes might
    cause one of the contestants to lose seats.

    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the :mod:`quota` module.
    '''
    def __init__(self,
                 quota_function: Union[str, Callable[[int, int], Number]],
                 ):
        self.quota_function = quota.construct(quota_function)
        self._quota_evaluator = QuotaDistributor(self.quota_function)

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute seats proportionally, rounding by largest remainder.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the proportional result
            awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        # first, assign the non-remainder seats for those over the quota
        quota_elected = self._quota_evaluator.evaluate(
            votes, n_seats, prev_gains
        )
        quota_number = self.quota_function(
            sum(votes.values()), n_seats
        )
        gained_prerem = util.sum_dicts(quota_elected, prev_gains)
        n_for_remainder = n_seats - sum(gained_prerem.values())
        remainders = {
            cand: Fraction(n_votes, quota_number) - gained_prerem.get(cand, 0)
            for cand, n_votes in votes.items()
            if gained_prerem.get(cand, 0) < max_seats.get(cand, INF)
        }
        for candidate in core.get_n_best(remainders, n_for_remainder):
            if candidate in quota_elected:
                quota_elected[candidate] += 1
            else:
                quota_elected[candidate] = 1
        return quota_elected


class HighestAverages:
    '''Distribute seats proportionally by ordering divided vote counts.

    Divides the vote count for each party by an increasing sequence of divisors
    (usually small integers), sorts these quotients and awards a seat for each
    of the first n_seats quotients.

    This includes some popular proportional party-list systems like D'Hondt or
    Sainte-Laguë/Webster The result is usually quite close to proportionality
    and avoids the Alabama paradox of largest remainder systems. However, it
    usually favors either large or smaller parties, depending on the choice
    of the divisor function.

    :param divisor_function: A callable producing the divisor from the number
        of seats awarded to the contestant so far. For example, the D'Hondt
        divisor (which uses the natural numbers sequence) would always return
        the number of currently held seats raised by one. The common divisor
        functions can be referenced by string name from the :mod:`divisor`
        module.
    '''
    def __init__(self,
                 divisor_function: Union[
                     str, Callable[[int], Number]
                 ] = 'd_hondt',
                 ):
        self.divisor_function = divisor.construct(divisor_function)

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute seats proportionally by highest averages.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds, to determine the starting divisor.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        totals = prev_gains.copy()
        quotient_dict = {}
        for cand, n_votes in votes.items():
            cand_total = totals.get(cand, 0)
            divisor = self.divisor_function(cand_total)
            if divisor > 0 and cand_total < max_seats.get(cand, n_seats):
                quotient_dict[cand] = Fraction(n_votes, divisor)
        candidates, quotients = [list(x) for x in zip(*util.sorted_votes(
            quotient_dict, descending=False
        ))]
        rem_seats = n_seats - sum(totals.values())
        while rem_seats > 0 and quotients:
            n_elect = 1
            max_q = quotients[-1]
            while n_elect < len(quotients) and quotients[-n_elect-1] == max_q:
                n_elect += 1
            to_elect = candidates[-n_elect:]
            if n_elect > rem_seats:
                to_elect = [core.Tie(to_elect)] * rem_seats
            for cand in to_elect:
                if cand in totals:
                    totals[cand] += 1
                else:
                    totals[cand] = 1
                rem_seats -= 1
            for i in range(n_elect):
                cand = candidates.pop()
                quotients.pop()
                cand_total = totals.get(cand, 0)
                if cand_total < max_seats.get(cand, n_seats) and cand in votes:
                    new_quot = Fraction(
                        votes[cand], self.divisor_function(cand_total)
                    )
                    new_pos = bisect.bisect_left(quotients, new_quot)
                    candidates.insert(new_pos, cand)
                    quotients.insert(new_pos, new_quot)
        return {
            cand: cand_seats - prev_gains.get(cand, 0)
            for cand, cand_seats in totals.items()
            if cand_seats > prev_gains.get(cand, 0)
        }


'''
class BiproportionalEvaluator:
    def __init__(self,
                 candidate_eval: DistributionEvaluator,
                 district_eval: Optional[DistributionEvaluator] = None,
                 ):
        self.candidate_eval = candidate_eval
        self.district_eval = district_eval if district_eval else candidate_eval
        self._agg = convert.ConstituencyVoteAggregator()

    def evaluate(self,
                 votes: Dict[Constituency, Dict[Candidate, int]],
                 n_seats: int,
                 ) -> Dict[Constituency, Dict[Candidate, int]]:
        # districts = list(votes.keys())
        # candidates = list(frozenset(
        #     cand for clist in votes.values() for cand in clist
        # ))
        district_sums = {
            district: sum(dvotes.values())
            for district, dvotes in votes.items()
        }
        district_alloc = self.district_eval.evaluate(district_sums, n_seats)
        cand_sums = self._agg.aggregate(votes)
        cand_alloc = self.candidate_eval.evaluate(cand_sums, n_seats)
        district_mult = {district: 1 for district in district_alloc.keys()}
        cand_mult = {cand: 1 for cand in cand_alloc.keys()}
        dist_cand_seats = {}
        print(district_alloc)
        print(cand_alloc)
        i = 1
        while True:
            print(i)
            prev_dist_cand_seats = dist_cand_seats
            # district-wise distribution: make sure each district has its
            # rightful amount of seats and distribute them among parties
            # accordingly
            dist_cand_seats = {}
            for district, dvotes in votes.items():
                part_votes = {
                    cand: cdvotes * cand_mult[cand] * district_mult[district]
                    for cand, cdvotes in dvotes.items()
                }
                dist_cand_seats[district] = self.candidate_eval.evaluate(
                    part_votes, district_alloc[district]
                )
            print(dist_cand_seats)
            if dist_cand_seats == prev_dist_cand_seats:
                print()
                print('END')
                print()
                return dist_cand_seats
            # adjust party vote multipliers
            cand_seats = self._agg.aggregate(dist_cand_seats)
            print(cand_seats)
            cand_mult = {
                cand: cand_alloc[cand] / cur_seats
                for cand, cur_seats in cand_seats.items()
            }
            print(cand_mult)
            # party-wise distribution: make sure each party has its rightful
            # amount of seats and distribute them among districts accordingly
            dist_seats = {}
            for cand, n_seats in cand_alloc.items():
                cvotes = {
                    district: (
                        votes[district].get(cand, 0)
                        * district_mult[district]
                        * cand_mult[cand]
                    )
                    for district in district_alloc.keys()
                }
                cand_eval = self.district_eval.evaluate(cvotes, n_seats)
                util.add_dict_to_dict(dist_seats, cand_eval)
            print(dist_seats)
            # adjust district vote multipliers
            district_mult = {
                district: district_alloc[district] / dseats
                for district, dseats in dist_seats.items()
            }
            print(district_mult)
            print()
            if i > 10:
                raise RuntimeError
            i += 1
'''



'''Evaluators for special partial purposes, especially tiebreaking.

These evaluators should not be used as the main component of an election
system (except for obscure ones). Many of them choose the winners randomly,
so they are useful for tiebreaking, but that is about it.

You can make the random evaluators outputs stable if you give them a seed for
the random generator, but be careful with that in a real-world setting.
'''

import random
from typing import Any, List, Dict, Optional
from numbers import Number

from ..candidate import Candidate
from .. import util


"""
class SpoiledBallotRemover:
    '''Returns only unspoiled ballots (votes for valid candidates).'''
    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 ) -> List[Candidate]:
        return [
            cand for cand in votes.keys()
            if not isinstance(cand, NoneOfTheAbove)
                and not isinstance(cand, ReopenNominations)
        ]
"""


class RandomUnrankedBallotSelector:
    '''Select candidates by drawing random simple ballots from the tally.

    Useful for tiebreaking. Can also be used to evaluate *random approval
    voting* if the approval votes are converted to simple first.

    :param seed: Seed for the random generator that performs the sampling.
    '''
    def __init__(self,
                 seed: Optional[int] = None,
                 ):
        self.seed = seed
        self.stable = (self.seed is not None)

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by drawing random ballots.

        :param votes: Simple votes.
        :param n_seats: Number of candidates (ballots) to be selected.
        '''
        random.seed(self.seed)
        return util.select_n_random(votes, n_seats)


class Sortitor:
    '''Perform sortition (random sampling) among the candidates.

    This selects the candidates purely randomly, with equal probabilities for
    each of them. Useful for tiebreaking and some obscure protocols such as
    Venetian doge election.

    :param seed: Seed for the random generator that performs the sampling.
    '''
    def __init__(self,
                 seed: Optional[int] = None,
                 ):
        self.seed = seed
        self.stable = (self.seed is not None)

    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates randomly.

        :param votes: Simple votes. The quantities of votes are disregarded.
        :param n_seats: Number of candidates to be selected.
        '''
        random.seed(self.seed)
        return util.select_n_random({
            cand: 1 for cand, n_votes in util.sorted_votes(votes)
        }, n_seats)


class InputOrderSelector:
    '''Select first N candidates as they appear in the vote counts.

    This is useful for tiebreaking with an externally determined sort order,
    e.g. by ballot numbers or pre-generated random numbers. It takes advantage
    of dictionaries in Python 3.7+ maintaining insertion order.
    '''
    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select the first candidates that appear in the votes dictionary.

        :param votes: Simple votes. The quantities of votes are disregarded.
        :param n_seats: Number of candidates to be selected.
        '''
        return [cand for i, cand in enumerate(votes.keys()) if i < n_seats]



'''Evaluators that operate sequentially on ranked votes.

This hosts mainly the transferable vote evaluator
(:class:`TransferableVoteSelector`) but also its less known relative,
:class:`PreferenceAddition`.
'''

import itertools
import collections
from fractions import Fraction
from typing import List, Dict, Tuple, Union, Callable, Optional
from numbers import Number

from ..candidate import Candidate
from ..vote import RankedVoteType
from ..component import transfer, quota
from . import core
from .. import util


class TransferableVoteSelector:
    '''Select candidates by eliminating and transfering votes among them.

    This is the evaluator for transferable vote selection (since the evaluator
    does not concern itself with restrictions on allowed votes, it covers not
    only single transferable vote systems - STV -, but also potential multiple
    transferable vote systems). Its single-winner
    variant is also called instant-runoff voting (IRV).

    First, this evaluator looks at first-preference votes. If any of the
    candidates has at least a specified quota of votes, they are elected and
    their votes over the quota are redistributed according to their next stated
    preference. If no candidate has the quota, the candidate with the fewest
    currently allocated votes is eliminated and their votes transferred
    according to their next stated preference. Votes that have no further
    stated preferences are called exhausted and are removed from consideration.

    The current version does not use elimination breakpoints.

    :param transferer: An instance determining how much votes for eliminated
        candidates to transfer and to whom. It must provide the
        :class:`transfer.VoteTransferer` interface (experimental, stability
        not guaranteed). The basic vote transfer variants are implemented
        in the :mod:`transfer` module and can be referred to by their names.
    :param retainer: A selector determining which candidates to retain when
        elimination is to be performed (it may accept a number of seats, which
        will correspond to the number of candidates to retain). If not given,
        the candidates with the lowest amounts of votes will be eliminated.
    :param eliminate_step: Determines how many candidates to eliminate whenever
        elimination is to be performed. If a negative integer, determines how
        many candidates to eliminate at each step (the default eliminates one
        candidate at a time). If a positive integer, determines how many
        candidates to retain after elimination - this essentially allows only
        a single elimination and might cause an infinite loop if not used
        properly.
    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the :mod:`quota` module. If not
        given, no election by quota takes place - the candidates are just
        eliminated until the desired number remains.
    :param accept_quota_equal: Whether to use non-strict comparison for the
        number of votes against the quota.
    '''
    def __init__(self,
                 transferer: Union[str, transfer.VoteTransferer] = 'Gregory',
                 retainer: Optional[core.Selector] = None,
                 eliminate_step: Optional[int] = -1,
                 quota_function: Union[
                     str, Callable[[int, int], Number], None
                 ] = None,
                 accept_quota_equal: bool = True,
                 # use_breakpoints: bool = True, - quota + applied
                 ):
        self.transferer = (
            getattr(transfer, transferer)() if isinstance(transferer, str)
            else transferer
        )
        if quota_function is not None:
            self.quota_function = quota.construct(quota_function)
        else:
            self.quota_function = None
        self.accept_quota_equal = accept_quota_equal
        self.retainer = retainer
        self.eliminate_step = eliminate_step

    def evaluate(self,
                 votes: Dict[RankedVoteType, Number],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by transferable vote.

        :param votes: Ranked votes. Equal rankings are allowed.
        :param n_seats: Number of candidates to select.
        '''
        return self.nth_count(votes, n_seats, 1000000000000)[1]

    def next_count(self,
                   allocation: Dict[Candidate, Dict[RankedVoteType, Number]],
                   quota: Number,
                   n_rem_seats: int,
                   ) -> Tuple[
                       Dict[Candidate, Dict[RankedVoteType, Number]],
                       List[Candidate]
                   ]:
        '''Advance the transferable voting process by one iteration (count).

        :param allocation: Current allocation of ranked votes to candidates
            still contesting the remaining seats.
        :param quota: Current value of the vote quota to elect a candidate.
        :param n_rem_seats: Number of seats not yet filled.
        :returns: A 2-tuple containing the new allocation of votes and a list
            of newly elected candidates (might be empty).
        '''
        totals = {
            cand: sum(cand_votes.values())
            for cand, cand_votes in allocation.items()
        }
        if len(totals) <= n_rem_seats:
            # all over count are eliminated, elect all remaining and end
            return {}, core.get_n_best(totals, len(totals))
        quota_elected = self._elect_by_quota(totals, quota, n_rem_seats)
        if quota_elected:
            return self.transferer.transfer(
                allocation,
                elected={cand: quota for cand in quota_elected},
            ), quota_elected
        else:
            # nobody elected by quota, we have to eliminate
            retained = self._select_retained(totals)
            newly_eliminated = [
                cand for cand in totals.keys()
                if cand not in retained
            ]
            return self.transferer.transfer(
                allocation, eliminated=newly_eliminated
            ), []

    def nth_count(self,
                  votes: Dict[RankedVoteType, Number],
                  n_seats: int = 1,
                  count_number: int = 1,
                  ) -> Tuple[
                      Dict[Candidate, Dict[RankedVoteType, Number]],
                      List[Candidate]
                  ]:
        '''Get the intermediate counting state at a given iteration (count).

        :param votes: Ranked votes. Equal rankings are allowed.
        :param n_seats: Number of candidates to select.
        :param count_number: 1-indexed count number.
        :returns: A 2-tuple containing the allocation of votes after the given
            count and a list of elected candidates so far (might be empty).
        '''
        allocation = self.first_preference_allocation(votes)
        new_allocation = None
        if self.quota_function:
            quota = self.quota_function(sum(votes.values()), n_seats)
        else:
            quota = float('inf')
        elected = []
        for count_i in range(count_number):
            if len(elected) >= n_seats:
                break
            if new_allocation is not None:
                allocation = new_allocation
            new_allocation, newly_elected = self.next_count(
                allocation, quota, n_seats - len(elected)
            )
            if not newly_elected and new_allocation == allocation:
                raise core.VotingSystemError('infinite loop in STV')
            elected += newly_elected
        return {
            cand: sum(cand_votes.values())
            for cand, cand_votes in allocation.items()
        }, elected[:n_seats]

    def _select_retained(self,
                         totals: Dict[Candidate, Number]
                         ) -> List[Candidate]:
        if self.retainer:
            if core.accepts_seats(self.retainer):
                n_seats = self._retained_count(totals)
                retained = self.retainer.evaluate(totals, n_seats)
            else:
                retained = self.retainer.evaluate(totals)
        else:
            retained = core.get_n_best(totals, self._retained_count(totals))
        if any(isinstance(e, core.Tie) for e in retained):
            raise NotImplementedError('tie in STV elimination')
        return retained

    def _retained_count(self, totals: Dict[Candidate, Number]) -> int:
        if self.eliminate_step is None:
            raise ValueError('need to specify eliminate step'
                             ' without standalone retainer')
        if self.eliminate_step < 0:
            return max(len(totals) + self.eliminate_step, 1)
        else:
            return min(self.eliminate_step, len(totals) - 1)

    def _elect_by_quota(self,
                        totals: Dict[Candidate, Number],
                        quota: Number,
                        n_rem_seats: int,
                        ) -> List[Candidate]:
        fill_quota = {
            cand: total for cand, total in totals.items()
            if total > quota or self.accept_quota_equal and total == quota
        }
        if fill_quota:
            if len(fill_quota) > n_rem_seats:
                new_elected = core.get_n_best(fill_quota, n_rem_seats)
                if any(isinstance(e, core.Tie) for e in new_elected):
                    raise NotImplementedError('tie in STV quota election')
            else:
                new_elected = core.get_n_best(fill_quota, len(fill_quota))
        else:
            new_elected = []
        return new_elected

    def first_preference_allocation(self,
                                    votes: Dict[RankedVoteType, int],
                                    ) -> Dict[
                                        Candidate, Dict[RankedVoteType, Number]
                                    ]:
        '''Allocate votes by first preference.

        Performed at the beginning of the first count.

        :param votes: Ranked votes.
        :returns: The votes dictionary separated into subdictionaries keyed by
            candidate to whom the votes are allocated.
        '''
        first_prefs = {
            cand: {} for cand in util.all_ranked_candidates(votes)
        }
        for vote, n_votes in votes.items():
            first_pref = vote[0]
            if isinstance(first_pref, collections.abc.Set):
                # shared rank
                # we pass this to transferer - we prepend a fictional
                # eliminated candidate to this vote and have it transferred
                first_prefs.setdefault(None, {})[(None,) + vote] = n_votes
            else:
                first_prefs[first_pref][vote] = n_votes
        if first_prefs.get(None):
            split_alloc = self.transferer.transfer(
                first_prefs, eliminated=[None],
            )
            for cand_alloc in split_alloc.values():
                to_remove = []
                to_update = {}
                for vote, n_votes in cand_alloc.items():
                    if vote[0] is None:
                        to_remove.append(vote)
                        to_update[vote[1:]] = n_votes
                for vote in to_remove:
                    del cand_alloc[vote]
                cand_alloc.update(to_update)
            return split_alloc
        else:
            return first_prefs


# Gradual preference addition
class PreferenceAddition:
    '''Evaluates ranked votes by stepwise addition of lower preferences.

    Each candidate starts with their first-preference votes and
    lower-preference votes are added to them until a sufficient amount of
    candidates achieve a majority. This includes the following voting systems:

    -   *Bucklin voting* where the lower preferences are added without change.
    -   *Oklahoma system* where the lower preferences are added divided by
        the order of preference, thus yielding the row of coefficients
        1, 1/2, 1/3, 1/4... (Use :class:`Fraction` to maintain
        exact values.)

    :param coefficients: Coefficients to multiply the lower preferences with
        before they are added to the vote totals (the default adds them simply,
        which creates Bucklin voting).
    :param split_equal_rankings: Whether to split votes having multiple
        alternatives at the same rank, or to add the whole amount to each
        of these alternatives.
    '''
    def __init__(self,
                 coefficients: Union[
                     List[Number],
                     Callable[[int], Number],
                 ] = [1],
                 split_equal_rankings: bool = True,
                 ):
        self.coefficients = coefficients
        self.split_equal_rankings = split_equal_rankings

    def evaluate(self,
                 votes: Dict[RankedVoteType, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by sequential preference addition.

        :param votes: Ranked votes.
        :param n_seats: Number of candidates to be elected.
        '''
        if self.split_equal_rankings:
            votes = self._decouple_equal_rankings(votes)
        total_votes = collections.defaultdict(int)
        majority_quota = Fraction(sum(votes.values()), 2)
        max_pref_len = max(
            len(preferences) for preferences, n_votes in votes.items()
        )
        elected = []
        for pref_i in range(max_pref_len):
            # add this round's preferences
            self._add_round_votes(total_votes, votes, pref_i, elected)
            # take all that have achieved majority, ordered by the vote sum
            majority = {
                cand: n_votes
                for cand, n_votes in util.sorted_votes(total_votes)
                if n_votes > majority_quota
            }
            best = core.get_n_best(majority, n_seats - len(elected))
            elected.extend(best)
            if len(elected) == n_seats:
                break
            else:
                for cand in best:
                    if cand in total_votes:
                        del total_votes[cand]
            # if still not enough elected, go on by adding another preference
        return core.Tie.reconcile(elected)

    def _decouple_equal_rankings(self,
                                 votes: Dict[RankedVoteType, int]
                                 ) -> Dict[RankedVoteType, Fraction]:
        new_votes = None
        for one_vote, n_votes in votes.items():
            equal_rank_tuples = [
                (i, rank) for i, rank in enumerate(one_vote)
                if isinstance(rank, collections.abc.Set)
            ]
            if equal_rank_tuples:
                if new_votes is None:
                    new_votes = votes.copy()
                del new_votes[one_vote]
                eqr_is = [i for i, rank in equal_rank_tuples]
                variants = list(itertools.product(*(
                    itertools.permutations(rank)
                    for i, rank in equal_rank_tuples
                )))
                n_variant_votes = Fraction(n_votes, len(variants))
                for variant in variants:
                    var_vote = one_vote[:]
                    offset = 0
                    for i, var_part in zip(eqr_is, variant):
                        var_vote = var_vote[:]
                        var_vote = (
                            var_vote[:i+offset]
                            + var_part
                            + var_vote[i+offset+1:]
                        )
                        offset += len(var_part)
                    new_votes[var_vote] = n_variant_votes
        return new_votes if new_votes is not None else votes

    def _add_round_votes(self,
                         total_votes: Dict[Candidate, Fraction],
                         votes: Dict[RankedVoteType, int],
                         pref_i: int,
                         elected: List[Candidate],
                         ) -> None:
        coef = self._get_coefficient(pref_i)
        for preferences, n_votes in votes.items():
            if pref_i < len(preferences):
                preference = preferences[pref_i]
                if isinstance(preference, collections.abc.Set):
                    n_to_add = n_votes * coef
                    for cand in preference:
                        if cand not in elected:
                            total_votes[cand] += n_to_add
                elif preference not in elected:
                    total_votes[preference] += n_votes * coef

    def _get_coefficient(self, pref_i: int) -> Number:
        if hasattr(self.coefficients, '__call__'):
            return self.coefficients(pref_i)
        elif pref_i < len(self.coefficients):
            return self.coefficients[pref_i]
        else:
            return self.coefficients[-1]
